/* $Id$
 *
 * python interface to libmad (the mpeg audio decoder library)
 *
 * Copyright (c) 2002 Jamie Wilkinson
 *
 * This program is free software, you may copy and/or modify as per the
 * GNU General Public License (version 2, or at your discretion, any later
 * version).  This is the same license as libmad.
 *
 * the code py_madfile_read was copied from the madlld program, which can be
 * found at http://www.bsd-dk.dk/~elrond/audio/madlld/
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "madmodule.h"
#include "pymadfile.h"

PyTypeObject py_madfile_t = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,
	"MadFile",
	sizeof(py_madfile),
	0,
	/* standard methods */
	(destructor) py_madfile_dealloc,
	(printfunc) 0,
	(getattrfunc) py_madfile_getattr,
	(setattrfunc) 0,
	(cmpfunc) 0,
	(reprfunc) 0,
	/* type categories */
	0, /* as number */
	0, /* as sequence */
	0, /* as mapping */
	0, /* hash */
	0, /* binary */
	0, /* repr */
	0, /* getattro */
	0, /* setattro */
	0, /* as buffer */
	0, /* tp_flags */
	NULL
};

/* functions */

PyObject * py_madfile_new(PyObject * self, PyObject * args) {
	py_madfile * mf;
	char * fname;
	char errmsg[512]; /* FIXME: use MSG_SIZE or something */
	unsigned long int bufsiz = 4096;
	int n;

	if (!PyArg_ParseTuple(args, "s|l:MadFile", &fname, &bufsiz)) {
		return NULL;
	}

	/* bufsiz must be an integer multiple of 4 */
#if 0
	switch (bufsiz % 4) {
		case 3: bufsiz--;
		case 2: bufsiz--;
		case 1: bufsiz--;
		default:
	}
#endif
	if ((n = bufsiz % 4)) bufsiz -= n;
	if (bufsiz <= 4096) bufsiz = 4096;
	printf("bufsiz is %ld\n", bufsiz);

	mf = PyObject_NEW(py_madfile, &py_madfile_t);
	if ((mf->f = fopen(fname, "r")) == NULL) {
		snprintf(errmsg, 512, "Couldn't open file: %s", fname);
		PyErr_SetString(PyExc_IOError, errmsg);
		PyObject_DEL(mf);
		return NULL;
	}
	printf("got the object\n");
	/* initialise the mad structs */
	mad_stream_init(&mf->stream);
	mad_frame_init(&mf->frame);
	mad_synth_init(&mf->synth);
	mad_timer_reset(&mf->timer);

	mf->framecount = 0;

	mf->buffy = malloc(bufsiz * sizeof(unsigned char));
	mf->bufsiz = bufsiz;

	return (PyObject *) mf;
}

static void py_madfile_dealloc(PyObject * self, PyObject * args) {
	if (PY_MADFILE(self)->f) {
		free(MAD_BUFFY(self));
		MAD_BUFFY(self) = NULL;
		MAD_BUFSIZ(self) = 0;
		fclose(PY_MADFILE(self)->f);
		mad_synth_finish(&MAD_SYNTH(self));
		/* for some reason, this function is segfaulting python when it cleans
		 * up */
		/* mad_frame_finish(&MAD_FRAME(self)); */
		mad_stream_finish(&MAD_STREAM(self));
	}
	PyObject_DEL(self);
}

/* convert the mad format to an unsigned short */
static unsigned short madfixed_to_ushort(mad_fixed_t f) {
	/* A fixed point number is formed of the following bit pattern:
	 *
	 * SWWWFFFFFFFFFFFFFFFFFFFFFFFFFFFF
	 * MSB                          LSB
	 * S = sign
	 * W = whole part bits
	 * F = fractional part bits
	 *
	 * This pattern contains MAD_F_FRACBITS fractional bits, one should
	 * always use this macro when working on the bits of a fixed point
	 * number.  It is not guaranteed to be constant over the different
	 * platforms supported by libmad.
	 *
	 * The unsigned short value is formed by the least significant whole
	 * part bit, followed by the 15 most significant fractional part bits.
	 * Warning: this is a quick and dirty way to compute the 16 bit number,
	 * madplay includes much better algorithms. */
	f = f >> (MAD_F_FRACBITS - 15);
	return (unsigned short) f;
}

/* this is a debugging macro, kill it off */
#define line() printf("%s: %d\n", __FUNCTION__, __LINE__)

static PyObject * py_madfile_read(PyObject * self, PyObject * args) {
	PyObject * buf, * tuple, * ret; /* output objects */
	char * buffy; /* output buffer */
	unsigned int i, size;
	int error = 0; /* did an error occur during this decode? */
	int nextframe = 0;

	/* nextframe might get set during this loop, in which case the bucket
	 * needs to be refilled */
	do {

	/* The input bucket must be filled if it becomes empty or if it's the
	 * first execution on this file */
	if ((MAD_STREAM(self).buffer == NULL) ||
		(MAD_STREAM(self).error == MAD_ERROR_BUFLEN)) {
		size_t readsize, remaining;
		unsigned char * readstart;

		/* [1] libmad may not consume all bytes of the input buffer.
		 *     If the last frame in the buffer is not wholly contained
		 *     by it, then that frame's start is pointed to by the
		 *     next_frame member of the mad_stream structure.  This
		 *     common situation occurs when mad_frame_decode():
		 *      - fails,
		 *      - sets the stream error to MAD_ERROR_BUFLEN, and
		 *      - sets the next_frame pointer to a non NULL value.
		 *     (See also the comment marked [2] below)
		 *
		 *     When this occurs, the remaining unused bytes must be put
		 *     back at the beginning of the buffer and taken into account
		 *     before refilling the buffer.  This means that the input
		 *     buffer must be large enough to hold a whole frame at the
		 *     highest observable bitrate (currently 448kbps).
		 */
		if (MAD_STREAM(self).next_frame != NULL) {
			remaining = MAD_STREAM(self).bufend - MAD_STREAM(self).next_frame;
			memmove(MAD_BUFFY(self), MAD_STREAM(self).next_frame, remaining);
			readstart = MAD_BUFFY(self) + remaining;
			readsize = MAD_BUFSIZ(self) - remaining;
		} else
			readsize = MAD_BUFSIZ(self),
			readstart = MAD_BUFFY(self),
			remaining = 0;

		/* Fill in the buffer.  If an error occurs, print a message
		 * and leave the decoder.  If the end of stream is reached
		 * we also leave the loop but the return status is left
		 * untouched.
		 */
		readsize = fread(readstart, 1, readsize, PY_MADFILE(self)->f);
		if (readsize <= 0) {
			if (ferror(PY_MADFILE(self)->f)) {
				fprintf(stderr, "read error: %s\n", strerror(errno));
				error = 1;
			}
			if (feof(PY_MADFILE(self)->f)) {
				fprintf(stderr, "end of input stream\n");
				/* clean up, exit */
				Py_INCREF(Py_None);
				return Py_None;
			}
		}

		/* Pipe the new buffer content to libmad's stream decode facility */
		mad_stream_buffer(&MAD_STREAM(self), MAD_BUFFY(self), readsize + remaining);
		MAD_STREAM(self).error = 0;
	}

	/* Decode the next mpeg frame.  The streams are read from the buffer,
	 * its constituents are broken down and stored in the frame structure,
	 * ready for examination/alteration or PCM synthesis.  Decoding options
	 * are carried to the Frame from the Stream.
	 *
	 * Error handling: mad_frame_decode() returns a non-zero value when an
	 * error occurs.  The error condition can be checked in the error member
	 * of the Stream structure.  A mad error is recoverable or fatal, the
	 * error status is checked with the MAD_RECOVERABLE macro.
	 *
	 * [2] When a fatal error is encountered, all decoding activities shall
	 *     be stopped, except when a MAD_ERROR_BUFLEN is signalled.  This
	 *     condition means that the mad_frame_decode() function needs more
	 *     input do its work.  Once should refill the buffer and repeat the
	 *     mad_frame_decode() call.  Some bytes may be left unused at the
	 *     end of the buffer if those bytes form an incomplete frame.  Before
	 *     refilling, the remaining bytes must be moved to the beginning of
	 *     the buffer and used for input for the next mad_frame_decode()
	 *     invocation.  (See the comment marked [1] for earlier details)
	 *
	 * Recoverable errors are caused by malformed bitstreams, in this case
	 * one can call again mad_frame_decode() in order to skip the faulty
	 * part and resync to the next frame.
	 */
	if (mad_frame_decode(&MAD_FRAME(self), &MAD_STREAM(self))) {
		if (MAD_RECOVERABLE(MAD_STREAM(self).error)) {
			fprintf(stderr, "recoverable frame level error: %s\n", mad_stream_errorstr(&MAD_STREAM(self)));
			fflush(stderr);
			/* go onto the next frame */
			nextframe = 1;
		} else {
			if (MAD_STREAM(self).error == MAD_ERROR_BUFLEN) {
				/* not enough data to decode */
				nextframe = 1;
			} else {
				fprintf(stderr, "unrecoverable frame level error: %s\n", mad_stream_errorstr(&MAD_STREAM(self)));
				return NULL;
			}
		}
	}
	
	} while (nextframe);

	/* The characteristics of the streams first frame are printed on stderr.
	 * The first frame is representative of the entire stream */
	if (PY_MADFILE(self)->framecount == 0) {
		/*
		if (printframeinfo(stderr, &MAD_FRAME(self).header)) {
			return NULL;
		}
		*/
	}
	
	/* Accounting.  The computed frame duration is in the frame header
	 * structure.  It is expressed as a fixed point number whose data type
	 * is mad_timer_t.  It is different from the fixed point format of the
	 * samples and unlike it, cannot be directly added or subtracted.  The
	 * timer module provides several functions to operate on such numbers.
	 * Be careful though, as some functions of mad's timer module receive
	 * their mad_timer_t arguments by value! */
	PY_MADFILE(self)->framecount++;
	mad_timer_add(&MAD_TIMER(self), MAD_FRAME(self).header.duration);

	/* Once decoded, the frame can be synthesised to PCM samples.
	 * No errors are reported by mad_synth_frame() */
	mad_synth_frame(&MAD_SYNTH(self), &MAD_FRAME(self));

line();
	/* Create the buffer to store the PCM samples in, so python can use it */
	size = MAD_SYNTH(self).pcm.length;
	printf("size = %d\nalloc = %d\n", size, size * sizeof(unsigned char));
	printf("buf = %p\n", buf);
line();

	buf = PyBuffer_New(size); /* sefgaulting here! */
	printf("buf = %p\n", buf);
line();
	tuple = PyTuple_New(1);
line();
	Py_INCREF(buf);
line();
	PyTuple_SET_ITEM(tuple, 0, buf);
line();
	if (!PyArg_ParseTuple(tuple, "t#", &buffy, &size)) {
		printf("no good\n");
line();
		return NULL;
	}
line();
	/* no longer use tuple */
	Py_DECREF(tuple);

line();
	/* die if we don't have the space */
	if (size < MAD_SYNTH(self).pcm.length) {
		fprintf(stderr, "hfs, the buffer is smaller than the PCM\n");
		return NULL;
	} else if (size == MAD_SYNTH(self).pcm.length) {
		printf("swooooit!\n");
	}

line();
	Py_BEGIN_ALLOW_THREADS

	/* Synthesised samples must be converted from mad's fixed point format
	 * to the consumer format.  Here we use unsigned 16 bit big endian ints
	 * on two channels.  Integer samples are temporarily stored in a buffer
	 * that is flushed when full. */
line();
	for (i = 0; i < MAD_SYNTH(self).pcm.length; i++) {
		unsigned short sample;

		/* left channel */
		sample = madfixed_to_ushort(MAD_SYNTH(self).pcm.samples[0][i]);
		*(buffy++) = sample >> 8;
		*(buffy++) = sample & 0xFF;

		/* right channel. 
		 * if the decoded stream is monophonic then the right channel
		 * is the same as the left one */
		if (MAD_NCHANNELS(&MAD_FRAME(self).header) == 2)
			sample = madfixed_to_ushort(MAD_SYNTH(self).pcm.samples[1][i]);
		*(buffy++) = sample >> 8;
		*(buffy++) = sample & 0xFF;
	}
line();

	Py_END_ALLOW_THREADS

line();
	printf("size is: %d\n", size);

line();
	ret = Py_BuildValue("Oi", buf, size);

	/* no longer use buf */
	Py_DECREF(buf);
line();

	/* pass back the PyBuffer object */
	return ret;
}

/* housekeeping */

static PyMethodDef madfile_methods[] = {
	{ "read",    py_madfile_read, METH_VARARGS, "" },
	{ NULL, 0, 0, NULL }
};

static PyObject * py_madfile_getattr(PyObject * self, char * name) {
	return Py_FindMethod(madfile_methods, self, name);
}
